/* MOXC -- a C version of Collinge's MOXIE language	*/

/*****************************************************************************
*		Change Log
*  Date		| Change
*-----------+-----------------------------------------------------------------
* 31-Dec-85 | Modified for use with midi
*  5-Feb-86 | Added m_rest and m_restuntil allowing rests at top level
* 28-May-86 | Added command line parsing
*  4-Jun-86 | Changed keyevent to separate calls for each event type
* 10-Jul-86 | Put loop in mainscore with prompt to play and replay
*  9-Oct-86 | JMaloney: Converted to Macintosh
* 27-Jan-87 | JMaloney: Converted to Lightspeed C
* 23-Mar-88 | Dannenberg: made cause delay an int to avoid int/long problem
*****************************************************************************/

/*
IMPORTS:
	asciievent(k)			user-defined action for terminal input
	bendchange(ch, val)		user-defined pitch bend handler
	ctrlchange(ch, c, val)	user-defined control change handler
	keydown(ch, p, v)		user-defined MIDI note on handler
	keyup(ch, p)			user-defined MIDI note off handler
	mainscore()				user-defined first action(s)
	peddown(ch)				user-defined pedal down handler
	pedup(ch)				user-defined pedal up handler
	prgmchange(ch, prgrm)	user-defined program change handler
	touchchange(ch, val)	user-defined aftertouch handler

EXPORTS:
	moxc.h
*/

#include "switches.h"
#include <storage.h>
#include <StdIO.h>
#include "cext.h"
#include "cmdline.h"
#include "midicode.h"
#include "moxc.h"
#include "mpu.h"
#include "userfns.h"
#include "userio.h"

/****************************************************************************
*
*	command line parsing
*
****************************************************************************/

#define nswitches 8
private char *switches[nswitches] =
	{"-debug", "-d", "-miditrace", "-m", "-trace", "-t", "-help", "-block"};

#define noptions 1
private char *options[1] = {"-tune"};

/****************************************************************************
*
*	defines and typedefs
*
****************************************************************************/

/* define SAFEMOXC to enable extra run-time checking */
#define SAFEMOXC true

/* number of events to prepare to run on each main loop */
#define NPREPARE 5

#define NPRIORITY 10

typedef struct event {
	struct event *next; /* link to next event record */
	ulong time;			/* time of this event */
	int priority;		/* priority of this event (0 is highest) */
	int (*routine)();	/* who to call */
	int p1, p2, p3, p4;	/* args to pass */
	int p5, p6, p7, p8;	/* more args to pass */
} *event_type;

/****************************************************************************
*
*	variables local to this module
*
****************************************************************************/

/*
 * pending is used to sort events by priority and to hold them until
 * there is time to insert them in the sorted evqueue.  note that
 * in this implementation, only one priority is used.
 * nodes are allocated from and returned to evfree.
 */

private boolean debug = false;		/* true to enable informative messages */
private event_type pending[NPRIORITY];	/* the pending queue (see note above) */
private int npending = 0;			/* how many events are in pending array */
private event_type evqueue = NULL;	/* waiting to run event queue */
private event_type evfree = NULL;	/* free list */
private boolean moxcdone;			/* flag to halt execution */
private int moxc_loudness = 127;	/* loudness of moxie "notes" and "pnotes" */
private  ulong eventtime; 			/* time of current event -- used to */
									/*	avoid timing errors due to finite
									 *	execution speed */

/****************************************************************************
*
*	routines private to this module
*
****************************************************************************/

void		cmdline_help(void);
event_type	evallocate(void);
event_type	evcdr(event_type *);
event_type	evcons(event_type, event_type);
void		evdeallocate(event_type);
boolean		evgtr(event_type, event_type);
event_type	evinsert(event_type, event_type);
void		evrun(void);
void		evshow(event_type);
void		moxcinit(int argc, char * argv[]);
void		moxcpoll(void);
void		schedule(void);

/****************************************************************************
*				cause
* Inputs:
*	int delay: time (in hundredths) before this event should occur
*	int (*routine)(): routine that implements the event
*	int p1 through p8: parameters to pass to routine
* Effect:
*	builds an event and puts it in pending queue for later scheduling
****************************************************************************/

public void cause(delay, routine, p1, p2, p3, p4, p5, p6, p7, p8)
	int delay;
	int (*routine)();
	int p1, p2, p3, p4, p5, p6, p7, p8;
{
	event_type ev;
	int priority = 0;

/* priorities are not currently implemented, but if they were:
	if (priority >= NPRIORITY) priority = NPRIORITY - 1;
	if (priority <= 0) priority = 0;
*/

#ifdef SAFEMOXC
	if (routine == 0) {
		printf("Error: cause called with NULL routine\n");
		musicterm();
		exit(1);
	}
#endif
    if (debug) printf("(Cause) eventtime is %lu\n", eventtime);
	ev = evallocate();
	ev->time = eventtime + delay;
	ev->priority = priority;
	ev->routine = routine;
	ev->p1 = p1;
	ev->p2 = p2;
	ev->p3 = p3;
	ev->p4 = p4;
	ev->p5 = p5;
	ev->p6 = p6;
	ev->p7 = p7;
	ev->p8 = p8;

	npending++;
	pending[priority] = evcons(ev, pending[priority]);
	if (debug) {
		printf("(cause) event is pending:\n");
		evshow(ev);
	}
}

/****************************************************************************
*				 cmdline_help
* Effect:
*	prints out command line help
****************************************************************************/

private void cmdline_help()
{
	fprintf(stderr,"program_name [options]\n");
	fprintf(stderr,"	 Options are:\n");
	fprintf(stderr,"	   -block           disable MIDI thru\n");
	fprintf(stderr,"	   -debug (-d)      enable verbose debug mode\n");
	fprintf(stderr,"	   -help            this message\n");
	fprintf(stderr,"	   -miditrace (-m)  turn on MIDI command trace\n");
	fprintf(stderr,"	   -tune file       use tuning from file\n");
	fprintf(stderr,"	   -trace (-t)      trace music\n");
}

/****************************************************************************
*				evallocate
* Returns:
*	event_type: new event allocated from freelist or with malloc
****************************************************************************/

private event_type evallocate()
{
	if (evfree) {
		return evcdr(&evfree);
	} else {
		return ((event_type) malloc(sizeof(struct event)));
	}
}

/****************************************************************************
*				evcdr
* Inputs:
*	event_type *evlist: address of a list pointer
* Returns:
*	event_type: the head of the list
* Effect:
*	removes the head of the list from *evlist
****************************************************************************/

private event_type evcdr(evlist)
	event_type *evlist;
{
	event_type temp = *evlist;

	*evlist = (*evlist)->next;
	temp->next = NULL;
	return temp;
}

/****************************************************************************
*				evcons
* Inputs:
*	event_type ev: event to put at the head of the list
*	event_type evlist: the current list
* Returns:
*	event_type: new list with ev at the head
* Effect:
*	modifies ev's next pointer
****************************************************************************/

private event_type evcons(ev, evlist)
	event_type ev;
	event_type evlist;
{
	ev->next = evlist;
	return ev;
}

/****************************************************************************
*				evdeallocate
* Inputs:
*	event_type ev: a node to deallocate
* Effect:
*	returns ev to free list
****************************************************************************/

private void evdeallocate(ev)
	event_type ev;
{
	evfree = evcons(ev, evfree);
}

/****************************************************************************
*				evgtr
* Inputs:
*	event_type ev1, ev2: two events to be compared
* Returns:
*	boolean: true if ev1 is before ev2
* Effect:
*	ev1 is "before" ev2 if it is earlier or has the same time
*	but a lower priority number than ev2
****************************************************************************/

private boolean evgtr(ev1, ev2)
	event_type ev1, ev2;
{
	return (ev1->time < ev2->time) ||
			((ev1->time == ev2->time) && (ev1->priority < ev2->priority));
}

/****************************************************************************
*				evinsert
* Inputs:
*	event_type evlist: the list (a priority queue)
*	event_type ev: the event to insert in evlist
* Returns:
*	event_type: the list resulting from inserting ev into evlist
* Implementation:
*	if ev goes at the head, ev is cons-ed onto evlist and returned
*	otherwise evlist is modified by inserting ev at the right spot
****************************************************************************/

private event_type evinsert(evlist, ev)
	event_type evlist;
	event_type ev;
{
	if (!evlist) {					/* no list, return event */
		ev->next = NULL;
		return ev;
	} else if (evgtr(ev, evlist)) {	/* make event first on list */
		return evcons(ev, evlist);
	} else {						/* insert event in proper place */
		register event_type ptr = evlist;

		while ((ptr->next) && evgtr(ptr->next, ev)) {
			ptr = ptr->next;
		}

		ev->next = ptr->next;
		ptr->next = ev;
		return evlist;
	}
}

/****************************************************************************
*				evrun
* Effect:
*	executes the previously scheduled event ev and deallocates it
****************************************************************************/

private void evrun()
{
	event_type ev;

	ev = evcdr(&evqueue);
	eventtime = ev->time;
	if (debug) {
		printf("(evrun) running an event:\n");
		evshow(ev);
	}

	(*(ev->routine)) /* apply the function to its arguments */
		(ev->p1, ev->p2, ev->p3, ev->p4, ev->p5, ev->p6, ev->p7, ev->p8);
	evdeallocate(ev);
}

/****************************************************************************
*				evshow
* Inputs:
*	eventtype ev: the event to show
* Effect:
*	prints a description of ev
* Assumes:
*	ev is not null
****************************************************************************/

private void evshow(ev)
	event_type ev;
{
	printf("address:    %lu\n", ev);
	printf("time:       %lu\n", ev->time);
	printf("priority:   %d\n",  ev->priority);
	printf("routine:    %d\n",  ev->routine);
	printf("parameters: %d, %d, %d, %d, %d, %d, %d, %d\n",
			ev->p1, ev->p2, ev->p3, ev->p4,
			ev->p5, ev->p6, ev->p7, ev->p8);
	askbool("Continue", true);
}

/****************************************************************************
*				loud
* Inputs:
*	int: dynamic: how loud (1 to 127, softest to loudest)
* Effect:
*	sets moxc_loudness
****************************************************************************/

public void loud(dynamic)
	int dynamic;
{
	moxc_loudness = dynamic;
}

/****************************************************************************
*					m_rest
* Inputs:
*	ulong time: amount of time to rest
* Effect:
*	waits until the amount of time specified has lapsed
* Assumes:
*	must not be called from a "caused" routine.  must only be called
*	from "mainscore" or a routine called directly or indirectly from
*	"mainscore" without using "cause".
****************************************************************************/

public void m_rest(time)
	ulong time;
{
	m_restuntil(time + gettime());
}

/****************************************************************************
*				  m_restuntil
* Inputs:
*	ulong time: event time to rest until
* Effect:
*	waits until the specified time has been reached (absolute time).
*	other "caused" events will take place during the rest provided
*	this routine is called from "mainscore" (see m_rest description).
****************************************************************************/
public void m_restuntil(time)
	ulong time;
{
	while(time > gettime()) {
		moxcpoll();
	}
}

/****************************************************************************
*				main
* Inputs:
*	int argc: number of command line arguments
*	char * argv: command line argument array
* Effect:
*	initializes and runs moxc program
****************************************************************************/

public void _main(argc,argv)
	int argc;
	char * argv[];
{
	while (askbool("Play", true)) {
		moxcinit(argc, argv);	/* initialize structures */
		mainscore();			/* call user's start program */
		while (!moxcdone) {
			if ((evqueue == NULL) && (npending == 0)) {
				/* we are done if:
				 *   there are no events on evequeue, AND
				 *   there are no pending events not yet scheduled
				 */
				moxcdone = true;
			}
			moxcpoll();			/* do work */
		}
		musicterm();
		printf("End of Moxc execution.\n");
	}
}

/****************************************************************************
*				moxcinit
* Inputs:
*	int argc: number of command line arguments
*	char * argv: command line argument array
* Effect:
*	initializes moxc system
****************************************************************************/

private void moxcinit(argc, argv)
	int argc;
	char * argv[];
{
	int i;

	cl_init(switches, nswitches, options, noptions, argv, argc);

	if (cl_switch("-help")) {
		cmdline_help();
		exit(0);
	}

	debug = (cl_switch("-debug") || cl_switch("-d"));

	for (i = 0; i < NPRIORITY; i++) {
		pending[i] = NULL;
	}
	npending = 0;
	evqueue = NULL;
	moxcdone = false;
	moxc_loudness = 127;
	eventtime = 0;
	musicinit();
}

/****************************************************************************
*				moxcpoll
* Effect:
*	dispatch on user inputs, cause events
****************************************************************************/

private void moxcpoll()
{
	ulong now;			/* current time */
	int ascii_key;		/* terminal input
						 *	(this is an int so users do not
						 *	 have to declare type of parameter) */
	byte midi_data[4];	/* midi input */

	/* get the time: */
	now = gettime();
	eventtime = now;

	/* see if any user-caused events have happened: */
	/* first, poll for and decode midi keyboard input: */
	if (getbuf(false, midi_data)) {
		byte code = midi_data[0] & MIDI_CODE_MASK;
		byte channel = (midi_data[0] & MIDI_CHN_MASK) + 1;
		if (code == MIDI_ON_NOTE) {
			if (midi_data[2] == 0) {	/* velocity 0 -> note off */
				keyup(channel, midi_data[1] - 12);
			} else {
				keydown(channel, midi_data[1] - 12, midi_data[2]);
			}
		} else if (code == MIDI_OFF_NOTE) {
			keyup(channel, midi_data[1] - 12);
		} else if (code == MIDI_TOUCH) {
			touchchange(channel, midi_data[1]);
		} else if (code == MIDI_BEND) {
			bendchange(channel, midi_data[1] + (midi_data[2] << 7));
		} else if ((code == MIDI_CTRL) && (midi_data[1] == SUSTAIN)) {
			if (midi_data[2] == 0) {	/* data 0 -> pedal up */
				pedup(channel);
			} else {
				peddown(channel);
			}
		} else if (code == MIDI_CTRL) {
			ctrlchange(channel, midi_data[1], midi_data[2]);
		} else if (code == MIDI_CH_PROGRAM) {
			prgmchange(channel, midi_data[1]);
		}
	}

	/* then, poll for ASCII keyboard event: */

	if (get_ascii(&ascii_key)) {
		asciievent(ascii_key);
		if (debug && (evqueue == NULL)) {
			printf("nextevent is scheduled for %lu\n", evqueue->time);
		}
	} else if (evqueue != NULL && now >= evqueue->time) {
		/* if it is now time, run the next event */
		evrun();
	}

	/* finally, schedule any events that are waiting to be scheduled */
	if (npending > 0) schedule();
}

/****************************************************************************
*				pnote
* Inputs:
*	int pitch: key number
*	int duration: duration in hundredths of a second
* Effect:
*	sound the given note and return immediately
****************************************************************************/

public void pnote(pitch, duration)
	int pitch;
	int duration;
{
	midi_note(1, pitch, moxc_loudness);
	cause(duration, (int (*)()) midi_note, 1, pitch, 0);
}

/****************************************************************************
*				note
* Inputs:
*	int pitch: key number
*	int duration: duration in hundredths of a second
* Effect:
*	sound the given note and don't return until it is over
****************************************************************************/

public void note(pitch, duration)
	int pitch;
	int duration;
{
	midi_note(1, pitch, moxc_loudness);
	m_rest((ulong) duration);
	midi_note(1, pitch, 0);
}

/****************************************************************************
*				quit
* Effect:
*	tells moxc to shut down
****************************************************************************/

public void quit()
{
	moxcdone = true;
}

/****************************************************************************
*				schedule
* Effect:
*	takes events off pending queues and prepares them to run
****************************************************************************/

private void schedule()
{
	register int events_prepared = 0;
	register int priority = 0;
	event_type ev;

	/* insert up to NPREPARE events from pending queues */
	while ((events_prepared < NPREPARE) && (priority < NPRIORITY)) {
		if (pending[priority]) {
			ev = evcdr(&pending[priority]);
			evqueue = evinsert(evqueue, ev);
			npending--;
			if (debug) {
				printf("(main) event inserted:\n");
				evshow(ev);
			}
			events_prepared++;
		} else {
			priority++;	/* do next priority */
		}
	}
}
