/* mm.c -- midi monitor; prints incoming midi commands */

/*****************************************************************************
*	    Change Log
*  Date	    | Change
*-----------+-----------------------------------------------------------------
*  7-Apr-86 | Created changelog
*  9-Oct-86 | JMaloney: Converted to Macintosh
* 26-Jan-87 | JMaloney: Converted to Lightspeed C
*****************************************************************************/

#include "switches.h"

#include <StdIO.h>
#include "cext.h"
#include "cmdline.h"
#include "midicode.h"
#include "midiparse.h"
#include "mpu.h"
#include "userio.h"

/****************************************************************************
*
*	command line switches and options
*
****************************************************************************/

#define nswitches 8
private char *switches[nswitches] = 
    {"-block", "-debug", "-d", "-help", "-miditrace", "-m", "-trace", "-t"};

#define noptions 1
private char *options[noptions] = {"-tune"};

/*****************************************************************************
*
*	routines local to this module
*
*****************************************************************************/

void	_main(int, char *[]);
void	output(byte[]);
void	put_pitch(int);
void	showhelp(void);

/****************************************************************************
*				main
* Inputs:
*	int argc: command line argument count
*	char *argv[]: command line argument array
* Effect:
*	prompts for parameters, starts monitor
****************************************************************************/

void _main(argc, argv)
    int argc;
    char *argv[];
{
	boolean accept_pitch_bends;
    byte data[4];
	int junk;

    cl_init(switches, nswitches, options, noptions, argv, argc);
    if (cl_switch("-help")) {
    	showhelp();
    	return;
    }

    accept_pitch_bends =
    	askbool("Monitor pitch bender and continuous controllers", false);

    musicinit();
    while (getbuf(false, data))
		/* clean out input buffer */;
    midi_cont(accept_pitch_bends);
    printf("Midi Monitor ready.	 Hit space to stop.\n");
    while (stop() == CONTINUE) {
		if (getbuf(false, data)) output(data);
    }
    musicterm();
    printf("\nMidi error word: %d\n", midi_error);
    printf("Done.\n");
}

/****************************************************************************
*				output
* Inputs:
*	byte data[]: midi data buffer holding one command
* Effect:
*	format and print a midi command
****************************************************************************/

private void output(data)
    byte data[];
{
    register int command;
    register int chan;

    command = data[0] & MIDI_CODE_MASK;
    chan = data[0] & MIDI_CHN_MASK;

    if ((command == MIDI_ON_NOTE) && (data[2] != 0)) {
		printf("Note On,  Chan %2d, Key %2d (",
			   chan, data[1]);
		put_pitch(data[1] - 12);
		printf("), Vel %d\n",
			   data[2]);
    } else if (((command == MIDI_ON_NOTE) && (data[2] == 0)) ||
    		    (command == MIDI_OFF_NOTE)) {
		printf("Note Off, Chan %2d, Key %2d (",
			   chan, data[1]);
		put_pitch(data[1] - 12);
		printf("), Vel %d\n",
			   data[2]);
    } else if (command == MIDI_CH_PROGRAM) {
		printf("Prog Chg, Chan %2d, Prog %2d\n",
			   chan, data[1]);
    } else if (command == MIDI_CTRL) {
		printf("Ctrl Chg, Chan %2d, Ctrl %2d, Val %2d\n",
			   chan, data[1], data[2]);
    } else if (command == MIDI_TOUCH) {
		printf("A. Touch, Chan %2d, Val %2d\n",
			   chan, data[1]);
    } else if (command == MIDI_BEND) {
		printf("P. Bend,  Chan %2d, Val %2d\n",
			   chan, (data[1] + (data[2] << 7)));
    } else {
    	printf("Unexpected command: %2u data: %2d, %2d, %2d\n",
    		   command, data[1], data[2], data[3]);
    }
}

/****************************************************************************
*				put_pitch
* Inputs:
*	int p: pitch number
* Effect:
*	write out the pitch name for a given number
****************************************************************************/

private void put_pitch(p)
    int p;
{
    static char *ptos[] = {
		"c", "cs", "d", "ef", "e", "f", "fs", "g", "gs", "a", "bf", "b"};

    printf("%s%d", ptos[p % 12], p / 12);
}

/****************************************************************************
*				showhelp
* Effect:
*	prints help text
****************************************************************************/

private void showhelp()
{
    fprintf(stderr,"MM is a midi monitor program; switches are:\n");
    fprintf(stderr,"   -block            disable MIDI thru\n");
    fprintf(stderr,"   -help             this message\n");
    fprintf(stderr,"   -miditrace (-m)   turn on MIDI command trace\n");
    fprintf(stderr,"   -trace (-t)       trace music\n");
}
