/* nointdriver.c
 *
 *	a replacement for mididriver.c that uses the Macintosh serial output
 *	driver (the "ROM" driver).  input is NOT supported!  this driver
 *	does NOT use interrupts.
 */

/*****************************************************************************
*		Change Log
*  Date		| Change
*-----------+-----------------------------------------------------------------
* 17-Jan-87 | JMaloney: Created
*****************************************************************************/

#include "switches.h"

#ifdef LIGHTSPEED
#include <StdIO.h>
#include <Errno.h>
#include <MacTypes.h>
#include <EventMgr.h>
#include <FileMgr.h>
#endif

#ifdef MPW
#include <StdIO.h>
#include <Errors.h>
#include <Types.h>
#include <Events.h>
#include <Files.h>
#endif

#include "cext.h"
#include "cmdline.h"
#include "midibuff.h"
#include "midicode.h"
#include "mididriver.h"
#include "mpu.h"
#include "userio.h"

/****************************************************************************
*
*	constants
*
****************************************************************************/

/* the modem port, also called port A */
#define portA 0
#define inPortA -6
#define outPortA -7

/* the printer port, also called port B */
#define portB 1
#define inPortB -8
#define outPortB -9

/* SCC base read & write addresses */
#define SCCRBase	0x1D8L
#define SCCWBase	0x1DCL

/* offsets for A & B control & data */
#define aData		0x6
#define aCtl		0x2
#define bData		0x4
#define bCtl		0x0

/****************************************************************************
*
*	routines private to this module
*
****************************************************************************/

void	init_driver(int);
void	scc_put(int, byte, byte);

/****************************************************************************
*					setupMIDI
* Inputs:
*	int port: driver (port) to initialize (either portA or portB)
*	int clockDivide: the clock divide setting, determined by the rate of
*		the external clock provided by your midi interface hardware:
*			0x40	500	kilohertz clock
*			0x80	  1	megahertz clock
*			0xC0	  2	megahertz clock
* Effect: 
*	initialized serial drivers and SCC chip for use with
*	the Opcode Systems "Studio Plus" hardware
****************************************************************************/

public void setupMIDI(port, clockDivide)
	int port;
	int clockDivide;
{
	ulong err;
	ulong junk;

#ifdef LIGHTSPEED
	char *driverAIn = "\P.AIn";
	char *driverAOut = "\P.AOut";
	char *driverBIn = "\P.BIn";
	char *driverBOut = "\P.BOut";
#endif

#ifdef MPW
	char *driverAIn = ".AIn";
	char *driverAOut = ".AOut";
	char *driverBIn = ".BIn";
	char *driverBOut = ".BOut";

	/* flush output before setting SCC due to MPW interrupt bug: */
	fflush(stdout);
	fflush(stderr);
#endif

	if (port == portA) {
		if ((err = OpenDriver(driverAIn, &junk)) != noErr)
			gprintf(ERROR,
				"error %ld opening modem port for input\n", err);
		if ((err = OpenDriver(driverAOut, &junk)) != noErr)
			gprintf(ERROR,
				"error %ld opening modem port for output\n", err);
	} else {
		if ((err = OpenDriver(driverBIn, &junk)) != noErr)
			gprintf(ERROR,
				"error %ld opening printer port for input\n", err);
		if ((err = OpenDriver(driverBOut, &junk)) != noErr)
			gprintf(ERROR,
				"error %ld opening printer port for output\n", err);
	}

	scc_put(port, 11, 0x28);
		/* make serial controller use external clock on TRxC */
	scc_put(port, 4, (clockDivide | 0x4));
		/* to set proper divide ratio for external clock */
	scc_put(port, 15, 8);
		/* disable all ext/stat but DCD */
}

/****************************************************************************
*					Xmit
* Inputs:
*	int port: port to send on
*	byte value: character to write
* Effect: 
*	writes given value to the appropriate serial output port
****************************************************************************/

public void Xmit(port, value)
	int port;
	byte value;
{
	ulong count = 1;
	byte buff[1];

	buff[0] = (char) value;
	FSWrite(
		((port == portA) ? outPortA : outPortB),
		&count, buff);
}

/****************************************************************************
*					restoreMIDI
* Inputs:
*	int port: port to restore
* Effect: 
*	writes given value to the appropriate serial output port
****************************************************************************/

public void restoreMIDI(port)
	int port;
{
	/* since ROM drivers are never closed, this is a noop */
}

/****************************************************************************
*					scc_put
* Inputs:
*	int port: 0 for modem port, 1 for printer port
*	byte reg: the scc register (0-15)
*	byte value: the value to write
* Effect: 
*	writes value into register "reg" of the scc chip for the given port
*	(see specs for Zilog Z8530 SCC Serial Communication Controller chip)
****************************************************************************/

private void scc_put(port, reg, value)
	int port;
	byte reg;
	byte value;
{
	register byte *addr;

	addr = (port == portA) ?
		*((byte **)SCCWBase) + aCtl :
		*((byte **)SCCWBase) + bCtl;
	*addr = reg;
	port = port;	/* noop, for timing */
	*addr = value;

/* gprintf(TRANS, "scc_put(%x, %x, %x, %lx)\n", port, reg, value, addr); */
	/* debugging note: this should print (for portA, the modem port):
	 *	scc_put(0, b, 28, bffffb)
	 *	scc_put(0, 4, 84, bffffb)
	 *	scc_put(0, f, 8, bffffb)
	 */
}
