/* noteoff.c -- keeps track of pending note off events for Adagio */

/*****************************************************************************
*		Change Log
*  Date		| Change
*-----------+-----------------------------------------------------------------
* 31-Dec-85 | Created changelog
*  2-Oct-86 | JMaloney: Converted to Macintosh
*  2-Oct-86 | JMaloney: Made memory allocation more efficient
*  7-Jan-87 | JMaloney: Converted to Lightspeed C
*****************************************************************************/

#include "switches.h"

#ifdef LIGHTSPEED
#include <StdIO.h>
#include <MemoryMgr.h>
#endif

#ifdef MPW
#include <StdIO.h>
#include <Types.h>
#include <Memory.h>
#endif

#include "cext.h"
#include "adagio.h"
#include "mpu.h"
#include "noteoff.h"
#include "userio.h"

/****************************************************************************
*
*	structures and variables private to this module
*
****************************************************************************/

/* off_type is a structure containing note-off information */

typedef struct off_struct {
	struct off_struct *next;
	ulong when;
	byte voice;
	char pitch;
} *off_type;

private off_type off_pool = NULL;		/* free list of off_type structures */
private off_type pending_offs = NULL;	/* active list */

/****************************************************************************
*
*	routines private to this module
*
****************************************************************************/

void	off_free(off_type off);

/****************************************************************************
*				note_offs
* Inputs:
*	ulong current_time: the current time
* Returns:
*	boolean: true if off list is not empty
* Effect:
*	turn off all notes whose time has come
* Implementation:
*	find and perform all note-off events in pending_offs which are
*	scheduled to occur at or before current_time
****************************************************************************/

public boolean note_offs(current_time)
	ulong current_time;
{
	register off_type temp;

	if ((pending_offs != NULL) && (pending_offs->when <= current_time)) {
		/* there is something to do */
		while ((pending_offs != NULL) && (pending_offs->when <= current_time)) {
			midi_note(pending_offs->voice, pending_offs->pitch, 0);
			temp = pending_offs;
			pending_offs = pending_offs->next;
			off_free(temp);
		}
	}
	return (pending_offs != NULL);
}

/****************************************************************************
*				off_free
* Inputs:
*	off_type off: a structure to deallocate
* Effect:
*	puts off on the freelist
****************************************************************************/

private void off_free(off)
	off_type off;
{
	off->next = off_pool;
	off_pool = off;
}

/****************************************************************************
*				off_init
* Effect:
*	initialize a pool of note-off records
****************************************************************************/

/* get 50 off_structs at a time */
#define USUAL_CHUNK (50L * sizeof(struct off_struct))

public void off_init()
{
	ulong desiredChunk = USUAL_CHUNK;
	ulong biggestChunk;
	long junk;
	off_type newOffs;
	int count;

	/* allocate space for a bunch of off_struct's */
	biggestChunk = MaxMem(&junk);
	desiredChunk = (biggestChunk < desiredChunk) ? biggestChunk : desiredChunk;
		/* if we can't get a chunk as large as we wanted, we'll take the
		 * biggest we can get */

	newOffs = (off_type) NewPtr(desiredChunk);
	if (newOffs == NULL) {
		gprintf(FATAL,
				"Out of memory for note offs. %lu bytes left.", FreeMem());
		clean_exit();
	}

	/* link new offs into the pool */
	count = desiredChunk / sizeof(struct off_struct);
	while (count > 0) {
		off_free(newOffs);
		newOffs++;
		count--;
	}
}

/****************************************************************************
*				off_schedule
* Inputs:
*	ulong offtime: time to turn note off
*	int voice: the midi channel
*	int pitch: the pitch
* Effect:
*	schedules a note to be turned off
* Assumes:
*	note_offs will be called frequently to actually turn off notes
* Implementation:
*	notes to be turned are kept in a time-ordered list; thus,
*	if there were ever a large number of notes in this list, inserting
*	another off_struct could be expensive
****************************************************************************/

public void off_schedule(offtime, voice, pitch)
	ulong offtime;
	int voice;
	int pitch;
{
	off_type off;
	register off_type  ptr;
	register off_type  prv;

	/* get a free off */
	if ((off = off_pool) == NULL) {
		/* out of offs */
		off_init();	/* make some more */
		if ((off = off_pool) == NULL) {
			/* should never happen */
			gprintf(FATAL, "Implementation error (noteoff.c): getting memory.");
			clean_exit();
		}
	}

	/* remove this off from the pool and initialize it */
	off_pool = off->next;
	off->when = offtime;
	off->voice = voice;
	off->pitch = pitch;

	/* insert into ordered list of pending note off events */
	ptr = pending_offs;
	if ((ptr == NULL) || (offtime <= ptr->when)) {
		/* put at front of list */
		off->next = ptr;
		pending_offs = off;
	} else {
		while ((ptr != NULL) && (offtime > ptr->when)) {
			/* traverse to right place in list */
			prv = ptr;
			ptr = ptr->next;
		}
		off->next = ptr;
		prv->next = off;
	}
}
