/* phase2.c -- plays an event list created by phase1 */

/*****************************************************************************
*		Change Log
*  Date		| Change
*-----------+-----------------------------------------------------------------
* 31-Dec-85 | Created changelog
*  2-Oct-86 | JMaloney: Converted to Macintosh
*  7-Oct-86 | JMaloney: Improved reporting when user stops
* 24-Oct-86 | JMaloney: Improved user interface; improved output of
*			|	simultaneous events (they are now all handled as a batch)
*  7-Jan-87 | JMaloney: Separated user interface from the basic play/record
*			|	loop to make it easier to use with different user interfaces.
*****************************************************************************/

#include "switches.h"

#ifdef LIGHTSPEED
#include <StdIO.h>
#endif

#ifdef MPW
#include <StdIO.h>
#endif

#include "cext.h"
#include "adagio.h"
#include "stream.h"
#include "midicode.h"
#include "mpu.h"
#include "noteoff.h"
#include "phase2.h"
#include "record.h"
#include "userio.h"

/****************************************************************************
*
*	constants
*
****************************************************************************/

/* WARNING: these must be long constants for Lightspeed C */

/* time offset from clock to performance */
#define ONE_SECOND 100L

/* latest possible time */
#define ETERNITY 0xFFFFFFFFL

/****************************************************************************
*
*	routines private to this module
*
****************************************************************************/

event_type	phase2_ProcessEvents(event_type, ulong, boolean);
void		phase2_Shutup(void);

/****************************************************************************
*				phase2_Play
* Inputs:
*	event_type score: score to play
*	boolean proceed: true if we are resuming an interrupted performance
*	boolean recording: true if we are recording as well as playing
*	ulong start_time: where to start playing
* Returns:
*	ulong *stop_time: set to time stopped
*	event_type *stop_place: set to next score event at stop time
*	int *stop_reason: set to indicate reason for stopping
* Effect:
*	plays the score from start_time, calling rec_poll periodically if
*	recording is true.
*	if proceed is true, set the clock to the current event time and
*	start immediately. (proceed is used for continuing after a pause.)
****************************************************************************/

#define SILENT true
#define NOISY false

void phase2_Play(
		score, proceed, recording, startTime,
		stop_time, stop_place, stop_reason)
	event_type score;
	boolean proceed;
	register boolean recording;
	ulong startTime;
	ulong *stop_time;
	event_type *stop_place;
	int *stop_reason;
{
	event_type nextEvent = score;	/* next event to perform */
	register ulong current_time;
	register boolean done = false;
	int stop_status;

	clear_abort();
	musicinit();
	if (!proceed) synth_init();

	if (proceed) {
		settime(nextEvent->ntime);
	} else if (startTime != 0) {
		nextEvent = phase2_ProcessEvents(score, startTime, SILENT);
		settime(startTime);
	} else {
		l_restuntil(ONE_SECOND);
			/* wait one second before beginning */
		timereset();
	}

	/* default return values */
	*stop_time = 0;
	*stop_place = score;
	*stop_reason = END_SCORE;

	/* notes about stopping:
	 *	when we are in play mode, we want to stop when we get to the end of
	 *	the piece.  when we are in record mode, we want to keep recording,
	 *	even after the entire score has been played, until the user wants
	 *	to stop.  the 'done' flag is used to detect the end of the
	 *	adagio score.
	 */
	while (!done) {		/* play it, Sam */
		current_time = gettime();
		if (!recording) done = true;
			/* assume we're done and see if this assumption is contradicted
			 * by finding more notes or note offs to process */

		if (recording) {
			/* record anything that's there */
			if (rec_poll(current_time)) {
				/* if rec_poll returns true, we are out of memory */
				phase2_Shutup();
				*stop_time = current_time;
				*stop_place = nextEvent;
				*stop_reason = OUT_OF_MEMORY;
				return;
			}
		}
		/* process note_offs first during any given tick to avoid
		 * gettin a ...note_off, note_on... sequence backwards
		 * and truncating a note.
		 */
		if (note_offs(current_time)) {
			done = false;	/* not done if there are more note offs */
		}
		if (nextEvent != NULL) {
			/* something to play */
			nextEvent = phase2_ProcessEvents(nextEvent, current_time, NOISY);
			done = false;	/* not done if there were notes to be played */
		}
		if ((stop_status = stop()) != CONTINUE) {
			/* user wants to stop */
			phase2_Shutup();
			*stop_time = current_time;
			*stop_place = nextEvent;
			*stop_reason = stop_status;
			return;
		}
	}
	musicterm();
}

/****************************************************************************
*				phase2_Shutup
* Effect:
*	turns off all notes currently sounding and calls musicterm
****************************************************************************/

private void phase2_Shutup()
{
	while (note_offs(ETERNITY))
		/* turn off all notes */;
	musicterm();
}

/****************************************************************************
*				phase2_ProcessEvents
* Inputs:
*	event_type event: a pointer to current place (event) in the score
*	ulong current_time: time when called
*	boolean silent: don't send note-on's
* Returns:
*	event_type: a pointer to the first event scheduled after current_time
*				  (i.e. the next event to be processed), or
*				NULL if there are no more events to be processed
* Effect:
*	processes all events with an event time at or before current_time.
*	if silent is true, note-on's are not peformed.  this feature is used
*	to jump to an arbitrary starting place in the score -- all sythesizer
*	state (programs, porta rate, etc) will be correctly set up since
*	control changes are output as usual.
* Caution:
*	it is possible that some midi devices may not be able to keep up with
*	a burst of messages, in which case a slight delay would have to be
*	introduced into the event sending loop...
****************************************************************************/

private event_type phase2_ProcessEvents(event, current_time, silent)
	register event_type event;
	register ulong current_time;
	register boolean silent;
{
	if (event == NULL) {
		return NULL;	/* nothing left to play */
	} else if (event->ntime > current_time) {
		return event;	/* nothing to do just now... */
	}

	while ((event != NULL) && (event->ntime <= current_time)) {
		/* process all current (and earlier) events */
		if (is_note(event)) {		/*** play a note or rest ***/
			/* set the program for this note */
			midi_program(event->nvoice, event->u.note.nprogram);
			/* if this note is not a rest, play it and schedule an off event */
			if ((event->u.note.npitch != NO_PITCH) && (!silent)) {
				midi_note(event->nvoice,
						  event->u.note.npitch,
						  event->u.note.nloud);
				off_schedule(event->ntime + event->u.note.ndur,
							 event->nvoice, event->u.note.npitch);
			}
		} else {					/*** send a control command ***/
			switch (event->ncontroller) {
				case 1:
					midi_ctrl(event->nvoice, PORTARATE, event->u.ctrl.value);
					break;
				case 2:
					midi_ctrl(event->nvoice, PORTASWITCH, event->u.ctrl.value);
					break;
				case 3:
					midi_ctrl(event->nvoice, MODWHEEL, event->u.ctrl.value);
					break;
				case 4:
					midi_touch(event->nvoice, event->u.ctrl.value);
					break;
				case 5:
					midi_ctrl(event->nvoice, FOOT, event->u.ctrl.value);
					break;
				case 6:
					midi_bend(event->nvoice, (event->u.ctrl.value << 6));
					break;
				default: break;
			}
		}
		event = event->next;
	}
	return event;
}
