/* record.c -- midi to adagio recorder
 *
 * the interface consists of these three routines:
 *
 *	boolean rec_init()	-- initialization
 *	boolean	rec_poll()	-- called during recording, returns true if
 *						   recording space is exhausted
 *	void	rec_final()	-- called to finish up
 */

/*****************************************************************************
*		Change Log
*  Date		| Change
*-----------+-----------------------------------------------------------------
* 27-Feb-86 | Created changelog
*			| Use pedal information when computing durations
* 23-Mar-86 | Determine size of transcription when rec_init is called
* 21-May-86 | Major rewrite to use continuous controls
*  9-Oct-86 | JMaloney: Converted to Macintosh
*  8-Jan-86 | JMaloney: Converted to Lightspeed C, made rec_final free buffer,
*			|	made rec_final write to a stream
*****************************************************************************/

#include "switches.h"

#ifdef LIGHTSPEED
#include <StdIO.h>
#include <MemoryMgr.h>
#endif

#ifdef MPW
#include <StdIO.h>
#include <Types.h>
#include <Memory.h>
#endif

#include "cext.h"
#include "cmdline.h"
#include "stream.h"
#include "midicode.h"
#include "mpu.h"
#include "record.h"
#include "userio.h"

/*****************************************************************************
* Implementation notes:
*	the midi stream is stored as an array of 4-byte records,
*	each of which is either a time or midi data.
*	midi data always begins with a control byte (high
*	order bit set), and it is assumed times have the high
*	order bit clear, so the two are easy to distinguish...
*	IF THE COMPILER PUTS THESE BITS IN THE SAME PLACE!
*	since it looks like the high order byte of the time lines up with
*	the last byte of a 4 byte array, we will always set the high order
*	bit of the last array byte when the first 3 bytes are filled
*	with MIDI data.  this is refered to as the "tag" bit.
*	WARNING: in some C compilers, longs are UNSIGNED, and thus
*	always positive.  test the high order bit with a mask!
*****************************************************************************/

#define MIDI_CMD		0x80
#define TAG_BIT			0x80000000L
#define istime(note)	(!(((note)->when) & TAG_BIT))
#define MAX_LINE		100

typedef union note_struct {
	byte n[4];
	ulong when;
} *note_type, note_node;

/* the event buffer */
private note_type event_buff;		/* pointer to allocated buffer */
private note_type next_event = NULL;/* pointer to next entry in buffer */
private note_type last_event = NULL;/* pointer to last entry in buffer */

/* table of pitch names */
private char *pitch_names[] = {
	"c", "cs", "d", "ef", "e", "f", "fs", "g", "gs", "a", "bf", "b"};

/* misc */
private boolean debug_rec = false;	/* verbose debugging */
private int max_pile_ups = 0;		/* maximum number of pile_ups */

/****************************************************************************
*
*	routines private to this module
*
****************************************************************************/

void		bend_filter(note_type, ulong);
void		byteorder(void);
void		ctrl_filter(note_type, ulong);
int			event_bend(note_type);
void		filter(void);
ulong		getdur(int, boolean, ulong);
ulong		getnext(int, ulong);
char		map_ctrl(int);
void		output(StreamPtr, boolean);

/****************************************************************************
*					bend_filter
* Inputs:
*	note_type current: the current event
*	ulong now: the current time
* Effect:
*	remove extra pitch bend events in same 0.01 sec time slot
* Implementation:
*	if there is another pitch bend in this time slot,
*	make the current event a no-op by replacing it with the time
* Assumes:
*	current is a pitch bend
****************************************************************************/

private void bend_filter(current, now)
	note_type current;
	ulong now;
{
	note_type future;

	/* look for additional pitch bends in this time slot */
	for (future = current + 1; future < last_event; future++) {
		if (istime(future) && (future->when > now)) {
			break;	/* new time slot */
		} else if (future->n[0] == current->n[0]) {
			current->when = now;
			return;	/* found another pitch bend in the current time slot */
		}
	}
}

/****************************************************************************
*					byteorder
* Effect:
*	check out assumptions about byte order and placement
****************************************************************************/

private void byteorder()
{
	if ((sizeof(event_buff[0]) != 4) ||
		(sizeof(event_buff[0].when) != 4) ||
		(sizeof(event_buff[0].n[0]) != 1)) {
			gprintf(FATAL, "Implementation error (record.c): size problem.");
			clean_exit();
	}
	event_buff[0].n[0] = 0x12;
	event_buff[0].n[1] = 0x34;
	event_buff[0].n[2] = 0x56;
	event_buff[0].n[3] = 0x78;
	if ((event_buff[0].when != 0x78563412) &&
		(event_buff[0].when != 0x12345678)) {
			gprintf(FATAL, "Implementation error (record.c): layout problem.");
			clean_exit();
	}
}

/****************************************************************************
*					ctrl_filter
* Inputs:
*	note_type current: the current event
*	ulong now: the current time
* Effect:
*	remove extra control change events in same 0.01 sec time slot
* Implementation:
*	if there is another change of this controller in this time slot,
*	make the current event a no-op by replacing it with the time
* Assumes:
*	current is a control change
****************************************************************************/

private void ctrl_filter(current, now)
	note_type current;
	ulong now;
{
	note_type future;

	/* look for additional changes of this controller in this time slot */
	for (future = current + 1; future < last_event; future++) {
		if (istime(future) && (future->when > now)) {
			break;	/* new time slot */
		} else if ((future->n[0] == current->n[0]) &&
				   (future->n[1] == current->n[1])) {
			current->when = now;
			return;	/* found another control change in the current time slot */
		}
	}
}

/****************************************************************************
*					event_bend
* Inputs:
*	note_type note: pointer to a pitch bend event
* Returns:
*	int: an 8 bit pitch bend number
****************************************************************************/

private int event_bend(note)
	note_type note;
{
	return (int) (((note->n[1]) >> 6) + ((note->n[2]) << 1));
}

/****************************************************************************
*					filter
* Effect:
*	allow only one control change per time slot (0.01 sec)
* Implementation:
*	call ctrl_filter and bend_filter to overwrite control changes with
*	noop data (the current time is used as a noop)
****************************************************************************/

private void filter()
{
	register note_type note;	/* loop control variable */
	ulong now;					/* last time seen */
	int command;				/* command pointed to by note */
	int chan;					/* channel pointed to by note */

	for (note = event_buff; note < last_event; note++) {
		if (istime(note)) {
			now = note->when;
		} else {
			command = note->n[0] & MIDI_CODE_MASK;
			chan = note->n[0] & MIDI_CHN_MASK;

			if ((command == MIDI_CTRL) &&
				(note->n[1] == SUSTAIN)) {
				/* do nothing */;
			} else if (command == MIDI_CTRL) {
				ctrl_filter(note, now);
			} else if (command == MIDI_TOUCH) {
				bend_filter(note, now);			/* bend and touch use the  */
			} else if (command == MIDI_BEND) {	/*  same filter routines   */
				bend_filter(note, now);
			}
		}
	}
}

/****************************************************************************
*					getdur
* Inputs:
*	int i: index of the note
*	boolean ped: true if pedal is down at event i
*	ulong now: the time at event i
* Returns:
*	ulong: the duration of note i
* Assumes:
*	event_buff[i] is a note
* Implementation:
*	this is tricky because of pedal messages.  the note is kept on by
*	either the key or the pedal.  we keep 2 flags, key and ped.  key is
*	turned off when a key is released, ped goes off and on with pedal.
*	note ends when (1) both key and ped are false or (2) key is
*	pressed (this event will also start another note).
****************************************************************************/

private ulong getdur(i, ped, now)
	int i;
	boolean ped;
	ulong now;
{
	note_type note;
	ulong start = now;
	int chan = event_buff[i].n[0] & MIDI_CHN_MASK;
	int pitch = event_buff[i].n[1];
	boolean key = true;	/* flag that says if note is on */
	boolean noteon;		/* true if a noteon message received on chan */
	boolean keyon;		/* true if noteon message had non-zero velocity */

	/* search from the next event to the end of the buffer: */
	for (note = &(event_buff[i+1]); note < last_event; note++) {
		if (istime(note)) {
			now = note->when;
		} else {
			noteon = keyon = false;
			if ((note->n[0] & MIDI_CHN_MASK) == chan) {
				noteon = ((note->n[0] & MIDI_CODE_MASK) == MIDI_ON_NOTE) &&
						  (note->n[1] == pitch);
				keyon = noteon && (note->n[2] != 0);
				if ((noteon && (note->n[2] == 0)) ||
					(((note->n[0] & MIDI_CODE_MASK) == MIDI_OFF_NOTE) &&
					 (note->n[1] == pitch)))
						key = false;
				if (((note->n[0] & MIDI_CODE_MASK) == MIDI_CTRL) &&
					(note->n[1] == SUSTAIN) && (note->n[2] == 127))
						ped = true;
				if (((note->n[0] & MIDI_CODE_MASK) == MIDI_CTRL) &&
					(note->n[1] == SUSTAIN) && (note->n[2] == 0))
						ped = false;
				if ((!key && !ped) || keyon)
						return now - start;
			}
		}
	}
	return last_event->when - start;
}

/****************************************************************************
*					getnext
* Inputs:
*	int i: the index of the current note
*	ulong now: the current time
* Returns:
*	ulong: the time of the next note, program, or control change
*		   (returns time of the last event if nothing else is found)
****************************************************************************/

private ulong getnext(i, now)
	int i;
	ulong now;
{
	note_type note;

	for (note = &(event_buff[i+1]); note < last_event; note++) {
		if (istime(note)) {
			now = note->when;
		} else {
			int cmd = note->n[0] & MIDI_CODE_MASK;

			if (((cmd == MIDI_ON_NOTE) &&
				(note->n[2] != 0)) ||		/* note on */
			   ((cmd == MIDI_CTRL) &&
				(note->n[1] != SUSTAIN)) ||	/* control change */
			   (cmd == MIDI_CH_PROGRAM) ||	/* program change */
			   (cmd == MIDI_TOUCH) ||		/* after touch */
			   (cmd == MIDI_BEND)) {		/* pitch bend */
				return now;
			}
		}
	}
	return last_event->when;
}

/****************************************************************************
*					map_ctrl
* Inputs:
*	int control: a midi control number
* Returns:
*	char: an adagio control change command letter, NULL if
*		  control change is not one of: PORTARATE, PORTASWITCH,
*		  MODWHEEL, FOOT
****************************************************************************/

private char map_ctrl(control)
	int control;
{
	switch (control) {
		case PORTARATE:		return 'J';
		case PORTASWITCH:	return 'K';
		case MODWHEEL:		return 'M';
		case FOOT:			return 'X';
		default:			return NULL;
	}
}

/****************************************************************************
*					output
* Inputs:
*	StreamPtr outStream: output stream
*	boolean absflag: true if first line of the adagio score should
*					 include the absolute time
* Effect:
*	translate contents of event_buff to adagio and output to outStream
* Implementation:
*	NOTE: put all program changes in rests
*	use N(ext) notation for all timing
*	output no more than one continuous controller parameter change
*	per clock tick for each continuous controller parameter
****************************************************************************/

private void output(outStream, absflag)
	StreamPtr outStream;
	boolean absflag;
{
	char lineBuff[MAX_LINE];	/* output line buffer */
	int i;						/* loop counter */
	int command;				/* the current command */
	int chan;					/* the midi channel of the current event */
	int lastchan = 0;			/* the default adagio channel (1) */
	boolean ped = false;		/* state of the pedal */
	int how_many =				/* number of events recorded */
		last_event - event_buff;
	ulong now;					/* the time of the this event */
	boolean bad_ctrl_flag		/* set to true if unknown ctrl read */
		= false;

	/* set the initial absolute time, all other times are relative */
	if (absflag) {
		sprintf(lineBuff, "t%lu ", event_buff[0].when);
		stream_PutString(outStream, lineBuff);
	}

	for (i = 0; i < how_many; i++) {
		if (debug_rec) {
			gprintf(TRANS,
				"event %d: %x %x %x (%lu)\n",
				 i, event_buff[i].n[0], event_buff[i].n[1],
				 event_buff[i].n[2], event_buff[i].when);
		}

		if (istime(event_buff+i)) {
			now = event_buff[i].when;
			if (debug_rec) gprintf(TRANS, "i = %d, now = %lu\n", i, now);
		} else {
			command = event_buff[i].n[0] & MIDI_CODE_MASK;
			chan = event_buff[i].n[0] & MIDI_CHN_MASK;

			if ((command == MIDI_ON_NOTE) && (event_buff[i].n[2] != 0)) {
				int midiKey = event_buff[i].n[1];
				sprintf(lineBuff, "%s%d u%lu l%d n%lu",
						pitch_names[midiKey % 12],
						(midiKey - 12) / 12,	/* adagio pitch = key - 12 */
						getdur(i, ped, now),
						event_buff[i].n[2],
						getnext(i, now) - now);
				stream_PutString(outStream, lineBuff);
				if (chan != lastchan) {
					sprintf(lineBuff, " v%d", chan + 1);
					stream_PutString(outStream, lineBuff);
					lastchan = chan;
				}
				stream_PutString(outStream, "\n");
			} else if (command == MIDI_CH_PROGRAM) {
				sprintf(lineBuff, "r z%d n%lu",
					event_buff[i].n[1] + 1,
					getnext(i, now) - now);
				stream_PutString(outStream, lineBuff);
				if (chan != lastchan) {
					sprintf(lineBuff, " v%d", chan + 1);
					stream_PutString(outStream, lineBuff);
					lastchan = chan;
				}
				stream_PutString(outStream, "\n");
			} else if ((command == MIDI_CTRL) &&
					   (event_buff[i].n[1] == SUSTAIN)) {
				ped = (event_buff[i].n[2] != 0);
			} else if (command == MIDI_CTRL) {
				char c = map_ctrl(event_buff[i].n[1]);
				if (c != NULL) {
					sprintf(lineBuff, "%c%d n%lu\n",
							c, event_buff[i].n[2],
							getnext(i, now) - now);
					stream_PutString(outStream, lineBuff);
				} else bad_ctrl_flag = true;
			} else if (command == MIDI_TOUCH) {
				sprintf(lineBuff, "O%d n%lu\n",
						event_buff[i].n[1],
						getnext(i, now) - now);
				stream_PutString(outStream, lineBuff);
			} else if (command == MIDI_BEND) {
				sprintf(lineBuff, "Y%d n%lu\n",
						event_bend(&event_buff[i]),
						getnext(i, now) - now);
				stream_PutString(outStream, lineBuff);
			} else if (command != MIDI_ON_NOTE) {
				gprintf(ERROR, "Command 0x%x ignored\n", command);
			}
		}
	}
	if (bad_ctrl_flag)
		gprintf(ERROR,
				"Some unrecognized control changes were omitted from file.\n");
}

/****************************************************************************
*					rec_final
* Inputs:
*	boolean absTime: true to output absolute time of the first note
* Effect:
*	write recorded data to outStream
****************************************************************************/

public void rec_final(outStream, absTime)
	StreamPtr outStream;
	boolean absTime;
{
	next_event->when = gettime();
	last_event = next_event;

	if (debug_rec)
		gprintf(TRANS, "max_pile_ups = %d\n", max_pile_ups);
	gprintf(TRANS, "%lu times and events recorded.\n",
			(ulong) (last_event - event_buff));
	filter();
	output(outStream, absTime);
	stream_Close(outStream);

	DisposPtr((Ptr) event_buff);
	event_buff = NULL;
}

/****************************************************************************
*					rec_init
* Inputs:
*	boolean recordBends: true if pitch bends should be recorded
* Returns:
*	boolean: true if initialization succeeds
* Effect:
*	prepares event_buff to record midi input
****************************************************************************/

#define ENOUGH_ROOM (20000L * sizeof(union note_struct))

public boolean rec_init(recordBends)
	boolean recordBends;
{
	ulong junk, biggestChunk, spaceForRecord, max_notes;

#ifndef APPLICATION
	debug_rec = (cl_switch("-debug") || cl_switch("-d"));
#endif

	byteorder();
	max_pile_ups = 0;

	/* allocate space, but only the first time rec_init() is called */
	if (event_buff == NULL) {
		biggestChunk = MaxMem(&junk);
		if (biggestChunk <= SPACE_FOR_PLAY) {
			/* not enough memory; give up */
			return false;
		} else {
			spaceForRecord =
				min((biggestChunk - SPACE_FOR_PLAY), ENOUGH_ROOM);
				/* leave SPACE_FOR_PLAY contiguous bytes of memory */
		}

		max_notes = spaceForRecord / sizeof(note_node);
		event_buff = (note_type) NewPtr(spaceForRecord);
		if (event_buff == NULL) {
			/* should never happen */
			gprintf(FATAL, "Implementation error (record.c): getting memory.");
			clean_exit();
		}
		next_event = event_buff;
		last_event = event_buff + (max_notes - 2);

		gprintf(TRANS,
			"Recording space for %lu events has been allocated.\n", max_notes);
	}
	while (getkey(false, (int *) &junk) != -1)
		/* flush old midi events */;

	midi_cont(recordBends);
	return (max_notes > 10);
		/* it would be silly to record with only room enough for 10 notes! */
}

/****************************************************************************
*					rec_poll
* Inputs:
*	ulong currTime: the current time
* Returns:
*	boolean: true if there is no more memory
* Effect:
	reads and stores any midi input since last call to rec_poll
* Assumes:
*	rec_poll must be called frequently to get accurate results
* Implementation:
*	time stamps and midi events share the same buffer of 4-byte events.
*	time is saved at most once per call to rec_poll and
*	 only if midi data is present.
****************************************************************************/

public boolean rec_poll(currTime)
	ulong currTime;
{
	int pile_ups = 0;

	/* note:
	 *	this code is a little "clever".  here's how it works:
	 *	we always store currTime in next_event->when, just
	 *	in case there *is* something in the input buffer.
	 *	then we attempt to fetch the next event into (++next_event->n).
	 *	if this succeeds, the following event, if any, goes into
	 *	(++next_event->n) and so forth and we finish by incrementing
	 *	next_event (in the loop test which fails) so that
	 *	next_event is left pointing to the next empty location.
	 *	otherwise, we back up next_event so that is is left pointing
	 *	to the slot where we initially stored currTime, ready for the next
	 *	call to rec_poll.
	 */

	next_event->when = currTime;
	while (getbuf(false, (++next_event)->n)) {
		next_event->n[3] = MIDI_CMD;	/* set tag bit on each event */
		pile_ups++;
		if (next_event >= last_event) {
			next_event++;	/* make it look like a failed getbuf */
			break;
		}
	}
	if (pile_ups == 0) {
		/* getbuf in loop test didn't get anything, so back up */
		next_event--;
	}

	if (pile_ups > max_pile_ups) max_pile_ups = pile_ups;
	return (next_event >= last_event);
}
