/* userio.c -- handy user interface functions; abort and stop handling */

/*****************************************************************************
*		Change Log
*  Date		| Change
*-----------+-----------------------------------------------------------------
* 21-May-86 | Created
*  4-Oct-86 | JMaloney: Converted to Macintosh
*  7-Oct-86 | JMaloney: Added clear_abort(), stop()
* 10-Oct-86 | JMaloney: Added get_ascii() for Moxc
* 22-Oct-86 | JMaloney: Moved abort handling to this module
*  7-Jan-87 | JMaloney: Converted to Lightspeed C
*  9-Jan-87 | JMaloney: Added "pause" feature to stop.
*****************************************************************************/

#include "switches.h"
#include "ToolboxUtil.h"
#include <StdFilePkg.h>
#include <FileMgr.h>

#ifdef LIGHTSPEED
#include <StdIO.h>
#include <CType.h>
#include <EventMgr.h>
#include <OSUtil.h>
#include <Strings.h>
#include <Unix.h>
#endif

#ifdef MPW
#include <StdIO.h>
#include <CType.h>
#include <Types.h>
#include <Strings.h>
#include <Events.h>
#include <OSUtils.h>
#include <Signal.h>
#endif

#include "cext.h"
#include "mpu.h"
#include "userio.h"

/****************************************************************************
*
*	variables shared with other modules
*
****************************************************************************/

public boolean aborted;
	/* exported variable; shared with macmidi.c
	 * set by abort_watcher();
	 * cleared by clear_abort();
	 * tested by stop()
	 */

/****************************************************************************
*
*	routines private to this module
*
****************************************************************************/

void	abort_watcher();
int GetReadFileName();
int GetWriteFileName();
void PtoC_StrCopy();

/****************************************************************************
*					abort_check
* Effect:
*	exit nicely if the aborted flag is set
****************************************************************************/

public void abort_check()
{
	if (aborted) clean_exit();
}

/****************************************************************************
*					abort_watcher
* Inputs:
*	SignalMap sigNumber: the signal number
*	SignalMap sigState: the state of all signals
*	SignalMap sigEnabled: which signals are currently enabled
*	(these are the inputs to any signal watcher; we don't use them)
* Effect:
*	sets the aborted flag when a command-period interrupt occurs
****************************************************************************/

#ifdef MPW
private void abort_watcher(sigNumber, sigState, sigEnabled)
	SignalMap sigNumber;
	SignalMap sigState;
	SignalMap sigEnabled;
{
	aborted = true;
}
#endif

/****************************************************************************
*					askbool
* Inputs:
*	char * prompt: string to prompt for user input
*	boolean expected: expected answer (true or false)
* Returns:
*	boolean: true or false as answered by user
* Effect:
*	prompts user for yes or no input and returns result
*	if user's answer differs from the expected answer, it asks for
*	confirmation
****************************************************************************/

#define undefined -1

public boolean askbool(prompt, deflt)
	char *prompt;
	boolean deflt;
{
	char answer[100];		/* user input string */
	char c;					/* user input */
	int result = undefined;	/* the result  */
	int sure;				/* confirmation */
							/*   -1 = undefined, 0 = false, 1 = true */

	while (result == undefined) {
		gprintf(TRANS, "%s? [%c]", prompt, (deflt ? 'Y' : 'N'));
		gets(answer);
		c = toupper(*answer);
		if (c == 'Y') {
			result = true;
		} else if (c == 'N') {
			result = false;
		} else if (c == 0)  {
			result = deflt;
		} else {
			gprintf(TRANS, "\nPlease type y or n.\n");
			continue;
		}
	}
	return result;
}

/****************************************************************************
*					clean_exit
* Effect:
*	clean up and exit
****************************************************************************/

public void clean_exit()
{
	gprintf(TRANS, "Exiting...\n");
	musicterm();
	exit(1);
}

/****************************************************************************
*					cleanup_abort_handler
* Effect:
*	shuts down abort watcher
****************************************************************************/

public void cleanup_abort_handler()
{
#ifdef MPW
	(void) sigset(SIGINT, SIG_DFL);	/* deactivate abort watcher */
#endif
}

/****************************************************************************
*					clear_abort
* Effect:
*	resets the aborted flag
****************************************************************************/

public void clear_abort()
{
	aborted = false;
}

/****************************************************************************
*					fileopen
* Inputs:
*	char * deflt: the default file name (e.g. from the command line)
*	char * extension: default extension
*	char * mode: read ("r") or write ("w")
*	char * prompt: prompt for user
* Returns:
*	FILE *: pointer to opened file
* Effect:
*	opens file, prompts for user input if necessary and warns about
*	possible confusion.  if deflt is a null string, the user will
*	be prompted for a name.	 the routine loops until a file is opened.
*	if the mode is "r", a check is made to see if the file exists
*	with and without the extension.	 if both exist, a warning is given.
*	for mode "w", a check is made to see if the file will be overwritten.
*	the extension is automatically added if the default or user-typed
*	file has no "."	 at the bottom of the loop body, if no file has
*	been opened, the user is prompted for another file name.
****************************************************************************/

public FILE * fileopen(deflt, extension, mode, prompt)
	char *deflt;
	char *extension;
	char *mode;
	char *prompt;
{
	char filename[100]; /* trial name */
	char extname[100];	/* trial name with extension added */
	FILE *fp = NULL;	/* file corresponding to filename */
	FILE *fpext;		/* file corresponding to extname */
	char *problem;		/* tells user why he has to try again */

	strcpy(filename, deflt);
	if (mode == NULL || mode[0] == 0)  return NULL;
	if (strlen(filename) > 0) {	/* if an initial filename was provided: */
		if (mode[0] == 'r') {
			strcpy(extname, filename);
			strcat(extname, ".");
			strcat(extname, extension);
			fp = fopen(filename, mode);
			fpext = fopen(extname, mode);
			if (fp != NULL && fpext != NULL) {
				gprintf(ERROR,
					"Warning: Both %s and %s exist.	 %s will be used.\n",
					filename, extname, filename);
				fclose(fpext);
			} else if (fpext != NULL) {
				fp = fpext;
			}
			if (fp == NULL) problem = "Couldn't find %s.";
		} else if (mode[0] == 'w') {
 			/* add the extension if there is no '.' in the file name */
			if (!strchr(filename, '.')) {
				strcat(filename, ".");
				strcat(filename, extension);
			}
			fp = fopen(filename, "r");
			if (fp != NULL) {
				char question[100];

				fclose(fp);
				strcpy(question, "OK to overwrite ");
				strcat(question, filename);
				if (askbool(question, false)) {
					fp = fopen(filename, mode);
				} else {
					fp = NULL;
					problem = "";
				}
			} else {
				fp = fopen(filename, mode);
				if (fp == NULL) problem = "Couldn't create %s.";
			}
		}
		if (fp == NULL) {
			gprintf(TRANS, problem, filename);
		}
	}
 	if (mode[0] == 'r') {
 		if (!GetReadFileName(filename)) return NULL;
		fp = fopen(filename, mode);
 	} else if (mode[0] == 'w') {
 		if (!(GetWriteFileName(filename, prompt))) return NULL;
 		/* add the extension if there is no '.' in the file name */
		if (!strchr(filename, '.')) {
			strcat(filename, ".");
			strcat(filename, extension);
		}
		fp = fopen(filename, mode);
 	} else {
 		gprintf(ERROR, "(fileopen) internal error: bad mode\n");
 	}
	return fp;
}

static int GetReadFileName(name)
char *name;
{
    static Point p = {100,100};
	SFReply loadfile;
    SFTypeList mytypes;
    
    mytypes[0] = 'TEXT'; 
/* could put any filter here (i.e. giofilefileter) */
    SFGetFile(p,"\p", NULL,1,mytypes,0L,&loadfile); 
    if (loadfile.good) {
		SetVol(0L,loadfile.vRefNum); 
	 	PtoC_StrCopy(&loadfile.fName, name);
	 	return(true);
	 }
	 else 
	 	return(false);
}


static int GetWriteFileName(fn, str)
char *fn, *str;
{
	static Point SFPwhere = { 106, 104 };
	char Pstr[100], Pfn[100];
 	SFReply reply;
 
 	strcpy(Pstr, str);
 	CtoPstr(Pstr);
 	strcpy(Pfn, fn);
 	CtoPstr(Pfn);

	SFPutFile(SFPwhere, Pstr, Pfn, 0L, &reply);
	if (reply.good) {
		SetVol (0L,reply.vRefNum);
		PtoC_StrCopy(&reply.fName, fn); 
		return(true);
	}
	else return(false);
}


/****************************************************************************
*					get_ascii
* Returns:
*	boolean: true if a character was found
*	int * c: pointer to int into which to store the character, if any
* Effect:
*	polls (doesn't wait) for an ascii character and says if it got one
*	the character is returned in *c.
****************************************************************************/

public boolean get_ascii(c)
	int *c;
{
	EventRecord theEvent;

	(void) GetNextEvent((keyDownMask | autoKeyMask), &theEvent);
	if ((theEvent.what == keyDown) || (theEvent.what == autoKey)) {
		*c = theEvent.message & charCodeMask;
		return true;
	} else {
		*c = -1;
		return false;
	}
}

/****************************************************************************
*					gprintf
* Inputs:
*	int * handler: pointer to output handler (say, a window)
*				 or one of {TRANS, ERROR, FATAL} from userio.h
*	char * format: a null-terminated printf style format string
*	int arg0 through arg14: a variable number of arguments for printf
* Effect:
*	formats and outputs the specified information to an output handler.
*	this is a machine and user-interface approach to output.  on
*	a simple machine, it is like printf.  on a more complex machine,
*	output is directed to the appropriate window.
****************************************************************************/

public void gprintf(
		where, format, arg0, arg1, arg2, arg3, arg4, arg5,
		arg6, arg7, arg8, arg9, arg10, arg11, arg12, arg13)
	Pointer where;
	char *format;
	int arg0, arg1, arg2, arg3, arg4, arg5, arg6;
	int arg7, arg8, arg9, arg10, arg11, arg12, arg13;
{
	char temp1[255];
	char temp2[255];

	sprintf(temp1, format,
			arg0, arg1, arg2, arg3, arg4, arg5, arg6,
			arg7, arg8, arg9, arg10, arg11, arg12, arg13);

	switch((long) where) {
	case TRANS:
		strcpy(temp2, temp1);
		break;
	case ERROR:
		strcpy(temp2, "ERROR: ");
		strcat(temp2, temp1);
		break;
	case FATAL:
		strcpy(temp2, "FATAL: ");
		strcat(temp2, temp1);
		strcat(temp2, "\n");
		break;
	default:
		strcpy(temp2, "UNKNOWN: ");
		strcat(temp2, temp1);
		strcat(temp2, "\n");
		break;
	}
	printf("%s", temp2);
}

/****************************************************************************
*					init_abort_handler
* Effect:
*	starts abort watcher
*	aborted flag is set to false
****************************************************************************/

public void init_abort_handler()
{
	aborted = false;						/* no aborts yet */
#ifdef MPW
	(void) sigset(SIGINT, abort_watcher);	/* activate abort watcher */
#endif
}


void PtoC_StrCopy(p1, p2)
  register char *p1, *p2;
/* copies a pascal string from p1 to p2 */
{
	register int len;
	
	len = *p1++;
	while (--len>=0) *p2++=*p1++;
	*p2 = '\0';
}


/****************************************************************************
*					stop
* Returns:
*	int: a stop code from {CONTINUE, STOP_REQ, PAUSE}
		 (these stop codes are defined in userio.h)
* Effect:
*	checks for a stop condition and returns an indication of which, if any
*	applies.  on the Macintosh, control-period is the usual abort key.
*	abort is tested by gettime() and causes a quiet death.
*	pressing the space bar will cause adagio to stop and print information
*	about the currently playing notes.
* Implementation:
*	checks Macintosh event queue for space key
*	NOTE: discards all non-space characters
****************************************************************************/

public int stop()
{
	int key;

	abort_check();
	while (get_ascii(&key)) {
		if (key == ' ') return STOP_REQ;
		else if ((key == 'p') || (key == 'P')) return PAUSE;
		/* else discard character and loop until event buffer empty */
	}
	return CONTINUE;
}

/****************************************************************************
*					wait_ascii
* Returns:
*	int: character for key pressed
* Effect:
*	waits for the user to type a key on the Macintosh terminal keyboard
*	(versus the synthesizer keyboard) and returns the key typed
****************************************************************************/

public int wait_ascii()
{
	int key;		/* key typed */
	long junk;		/* not examined */

	fflush(stdout);	/* make sure any prompts or errors have been output */
	fflush(stderr);
#ifndef LIGHTSPEED
	while (!get_ascii(&key)) {
		/* wait 1/10 second before retrying */
		Delay(6, &junk);
	}
#else
    do {
    	char input[100];
    	gets(input);
    	key = *input;
    } while (key == 0);
#endif
	return key;
}
